/*

	Simple program for Quantum Circuit Matrix calculation 
	It outputs som many matrices as are segments in the circuit
	

*/




#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <String>
#include <complex>
#include <iostream>
#include <fstream>
#include <strstream>

//#define __STD_COMPLEX

using namespace std;


const int MAXNUMOFGATES = 20;		// number of gates used
const int MAXGATEDEM = 32;			// 2 to the n power, here set n = 5, so 2^5 = 32. also max number of individual matirx size
const int MAXINDVWIRENUM = 5;		// max number of wires for input of an individual is MAXINDIWIRENUM
const int MAXSEGNUM = 20;			// max number of segments per individual is MAXSEGNUM
const int MAXSTRING = 150;			// max number of chars in an individual string
const int MAXGATEINPUT = 5;			// max number of input for any single gate

typedef complex<double> dComp;

struct tempT
{
       int numberGates;					// number of gates in this catagory
       char nameGate[MAXNUMOFGATES];	// for a particular number of input and output, the list of the gates can be used
};

struct qGate
{
       char representation;							// using a character to represent it
       int numInput;								// number of input wires
       int numOutput;								// number of output wires
	   int Cost;									// cost 
       dComp gateMatrix1[MAXGATEDEM][MAXGATEDEM];	// matrix of this gate
};

struct Individual
{
       int inputNumber;
       int outputNumber;
       int segmentNumber;
       int numberofgates;
       string my_string;
       float fitness;
	   //for either complex fitness function 
	   //either pareto-optimality
	   float Cost;
	   float Error;
	   //number of circuits i was used for
	   int Groups;
};


// holds GA functions
class GA
{
public:
  GA(){};
  ~GA(){};
       //* Set complex double variable equal to complex double constant */
  dComp I; //complex variable


  void Initialize(); // initialize the population
  void individualMatrix(bool);

  void closeStream();
  void output(int);
  Individual Indv;
private:
  qGate gateArray[MAXNUMOFGATES];				// an array of object qGate
  qGate nullGate;								//wire gate
  qGate finalGate;								// optimal result gate
  tempT stringArray[MAXINDVWIRENUM];			// sorted by input wires in asending order

  qGate tensorProduct(qGate, qGate);			// tensor product of gate A and gate B
  qGate matrixProduct(qGate, qGate);			// matrix product of gate A and gate B
  int getGate(char);
  int numofgates;								// number of gates read from input file
  int resultnum;								// the number of wire for input and output of the result gate

  ifstream in_stream;							// input file stream
  ofstream out_stream;							// output file stream

  int generationCondition;						// termination criterian
  bool condition;
  void initializeGates();
  void setStringArray();
};

int main()
{

	srand( time(NULL) );
	int generationCondition = 0;

	//ignition
	GA g;

	//population ignition
	g.Initialize();

	g.individualMatrix(false);


	//---------printing---------------
	g.output(generationCondition);
	g.closeStream();
	return 0;
}


void GA::individualMatrix(bool display)
{    
	//temporary storage values
	float Error = 0;
	int Time = 0;
	float Cost = 0;
	
	//three free gates for operations
	//finalG is the mtrix represenation for fitness evaluation
	qGate temp0, temp1, temp2, finalG;
	if (Indv.inputNumber <= 1) 
	{  
		int j = 0;
		for (int a = 0;a<= MAXNUMOFGATES;a++)
			if (gateArray[a].representation == Indv.my_string.at(j))
			{
				temp0 = gateArray[a];
				Cost+= gateArray[a].Cost;
				break;
			}
		j++;
		
		while (j<=Indv.my_string.length()-1)
		{
			for ( a = 0;a<= MAXNUMOFGATES;a++)
				if (gateArray[a].representation == Indv.my_string.at(j))
				{
					temp2 = gateArray[a];
					Cost+= gateArray[a].Cost;
					break;
				}
			
			temp0 = matrixProduct(temp0, temp2);
			j++;
		}
		finalG = temp0;
	} else //for more than one wire circuit
	{
		int begin = 0;
		int end = Indv.my_string.find("p", begin+1);

		while(end > 0) {
			if ((end - begin) > 1)
			{
				for (int a = 0;a<= MAXNUMOFGATES;a++)
					if (gateArray[a].representation == Indv.my_string.at(begin+1)) 
					{
						temp0= gateArray[a]; 
						begin++; 
						Cost+= gateArray[a].Cost; 
						break;
					}
				for (int b = begin+1;b<end;b++) 
				{
					for (int a = 0;a<= MAXNUMOFGATES;a++)
						if (gateArray[a].representation == Indv.my_string.at(b)) 
						{
							temp1= gateArray[a]; 
							Cost+= gateArray[a].Cost; 
							break;
						}
						
						

					temp0 = tensorProduct(temp0, temp1);
				}
				if (Time == 0) 
				{
					finalG = temp0;
					Time++;
				}
				else 
				{
					finalG = matrixProduct(finalG, temp0); 
					Time++;
				}

				// displaying intermediary gates
				if (display)
				{
					out_stream << " Segment number " <<Time<<endl;
					for (int a = 0; a<(int(pow(2,finalG.numInput)));a++)
					{
						for (int b = 0; b<(int(pow(2,finalG.numInput)));b++)
							out_stream << finalG.gateMatrix1[a][b]<<" ";
						out_stream << endl;
					}
				}
			}
			begin = Indv.my_string.find("p", end);
			end = Indv.my_string.find("p", begin+1);
		}
	}
	finalG.numInput = Indv.inputNumber;
	
	//for pareto
	Indv.Cost = Cost;
	
	if (display)
	{
		out_stream << " -------- " <<endl;
		for (int a = 0; a<(int(pow(2,finalG.numInput)));a++)
		{
			for (int b = 0; b<(int(pow(2,finalG.numInput)));b++)
				out_stream << finalG.gateMatrix1[a][b]<<" ";
			out_stream << endl;
		}
	}
}

void GA::setStringArray()
{
     int count[MAXNUMOFGATES];
     for (int j = 0; j < MAXGATEINPUT; j++)
     {
         count[j] = 0;
         for (int i = 0; i < numofgates; i++)
         {
              if (gateArray[i].numInput == (j+1))
              {
                 stringArray[j].nameGate[count[j]] = gateArray[i].representation;
                 count[j]++;
              }
              stringArray[j].numberGates = count[j];
         }
     }
}


void GA::output(int generationCondition)
{
     out_stream << "There are " << numofgates << " gates from the input file.\n";
     for (int i = 0; i < numofgates+1; i++)
     {
         out_stream << "The gate number " << i << " is: " << endl;
         out_stream << "And the gate representation is: " << gateArray[i].representation << endl;
         out_stream << "Number of inputs is: " << gateArray[i].numInput << endl;
         for (int j = 0; j < (int(pow(2,gateArray[i].numInput))); j++)
         {
             for (int k = 0; k < (int(pow(2,gateArray[i].numInput))); k++)
             {
                 out_stream << gateArray[i].gateMatrix1[j][k] << " ";
             }
             out_stream << endl;
         }
     }


      out_stream << "---------------------------" << endl;
      out_stream << " The resulting matrix of this ircuit is " << endl;
      out_stream << "---------------------------" << endl;
      out_stream << "----- Individual -----" << endl;

	  out_stream << Indv.inputNumber<<" number of inputs "<<endl;
	  individualMatrix(true);
	  out_stream<<endl;
	  out_stream<< Indv.my_string<<" representation "<<endl;
	  out_stream<< Indv.Cost<<" Cost "<<endl;
}


// Restrictions:
// 1. no fan out - ex, 5 inputs has to have 5 outputs
// 2. maximum 5 I/O (geometric increase of complexity with more I/Os)
void GA::Initialize()
{
     char inFile[100];
     char outFile[100];
     condition= false;

     // Open input and output files.
     cout << "Please input the source file" << endl;
     cin >> inFile;


     in_stream.open(inFile);
     if (in_stream.fail())
     {
        cout << "Input file opening failed.\n";
        exit(1);
     }

     cout << "\nPlease input the output file" << endl;
     cin >> outFile;

     out_stream.open(outFile);
     if (out_stream.fail())
     {
        cout << "Output file opening failed.\n";
        exit (1);
     }

     initializeGates();
     setStringArray();

 
}


int GA::getGate(char b)
{
	for (int a = 0;a<MAXNUMOFGATES;a++)
		if (gateArray[a].representation == b) return a;
  
}





void GA::initializeGates()
{
	I = dComp (0., 1.); //  Declare complex double variables
	double x;
	int counter = 0;

	in_stream>>Indv.inputNumber;
	Indv.outputNumber = Indv.inputNumber;
	in_stream>>Indv.my_string;

    in_stream >> numofgates;

	nullGate.numInput = 1;
	nullGate.numOutput = 1;
	nullGate.representation = 'W';
	nullGate.gateMatrix1[0][0] = 1;
	nullGate.gateMatrix1[0][1] = 0;
	nullGate.gateMatrix1[1][0] = 0;
	nullGate.gateMatrix1[1][1] = 1;
	//cost of an empty wire
	nullGate.Cost = 0;
	gateArray[0] = nullGate;

	out_stream<< gateArray[0].representation <<endl;
	out_stream<<"  Cost of the gate  "<< gateArray[0].Cost <<endl;

	for (int j = 0; j < (int(pow(2,gateArray[0].numInput))); j++)
      {
            for (int k = 0; k < (int(pow(2,gateArray[0].numInput))); k++)
            {out_stream<< gateArray[0].gateMatrix1[j][k]<<' ';
		    }
		out_stream<< endl;
      }


     // get first gate
    for (int a = 1; a < numofgates+1; a++)
     {
         in_stream >> gateArray[a].numInput;
         gateArray[a].numOutput = gateArray[a].numInput;
         gateArray[a].representation = char(a+65); // starting from A

		 in_stream >> gateArray[a].Cost;

		 out_stream<< gateArray[a].representation <<endl;
		 out_stream<<"  Cost of the gate  "<< gateArray[a].Cost <<endl;

         for (int j = 0; j < (int(pow(2,gateArray[a].numInput))); j++)
         {
             for (int k = 0; k < (int(pow(2,gateArray[a].numInput))); k++)
             {
				in_stream >> x;
				out_stream<< x <<' ';
                 if (x == 2)
                  gateArray[a].gateMatrix1[j][k] = I;
                 else if (x == -2)
                         gateArray[a].gateMatrix1[j][k] = -I;
                      else if (x == 4)
							gateArray[a].gateMatrix1[j][k] = exp(complex<double>(0, (3.1416/4)));
							else if (x == 5)
								gateArray[a].gateMatrix1[j][k] = complex<double>(0.5, 0.5);
								else if (x == -5)
									gateArray[a].gateMatrix1[j][k] = complex<double>(0.5, -0.5);
									else gateArray[a].gateMatrix1[j][k] = x;
				out_stream<< gateArray[a].gateMatrix1[j][k]<<' ';
             }
			out_stream<< endl;
         }
      counter++;
     }
   
}


qGate GA::tensorProduct(qGate A, qGate B)
{

      qGate productGate;
      for (int i = 0; i < (int(pow(2,A.numInput))); i++)
      {    for (int j = 0; j < (int(pow(2,A.numInput))); j++)
           {   for (int k = 0; k < (int(pow(2,B.numInput))); k++)
               {   for (int l = 0; l < (int(pow(2,B.numInput))); l++)
                   {  
                      productGate.gateMatrix1[i*(int(pow(2,B.numInput)))+k][j*(int(pow(2,B.numInput)))+l] = A.gateMatrix1[i][j] * B.gateMatrix1[k][l];
                   }
              }

          }
      }


      productGate.numInput = (A.numInput + B.numInput);
      productGate.numOutput =(A.numInput + B.numInput);

      return productGate;
}


qGate GA::matrixProduct(qGate A, qGate B)
{ qGate producGate;
       //rows of the first gate
      for (int i = (int(pow(2,A.numInput))-1);i>=0; i--)
      {   //colums of the second gate
          for (int j = 0; j < (int(pow(2,B.numInput))); j++)
          {   //colums of the first matrix
		int l = (int(pow(2,A.numInput))); 
              for (int k = 0; k < (int(pow(2,A.numInput))); k++)
              { //rows of the second
producGate.gateMatrix1[j][i] = producGate.gateMatrix1[j][i] + A.gateMatrix1[k][i] * B.gateMatrix1[j][k];
 		l--;
              }
          }
      }
      producGate.numInput = A.numInput;
      producGate.numOutput = producGate.numInput;
      return producGate;

}

void GA::closeStream()
{
     in_stream.close();
     out_stream.close();
}


